from enum import Enum
import numpy as np
from PythonCoordinates.measurables.physical_quantities import Length, Angle, Speed, Temperature, Humidity, Pressure
from PythonCoordinates.coordinates.coordinate_representations import SphericalCoordinate


class AircraftType(Enum):
    Military = 0, 'Military'
    Civilian = 1, 'Civilan'


class DataCollectionType(Enum):
    Measured = 0, 'Measured'
    Estimated = 1, 'Estimated'


class InterpolationCode(Enum):
    Fixed = 0, 'Fixed'
    Parallel = 1, 'Parallel'
    Variable = 2, 'Variable'


class NoiseFilePowerSetting:
    def __init__(self):
        self.value = 0.0
        self.units = ''
        self.upper_limit = 0.0
        self.lower_limit = 0.0


class NoiseFileAcousticSource:

    def __init__(self, filename=None):
        super().__init__()
        # in the C# code this is an observable collection. Not sure why...
        self._power_settings = []

        self._vehicle_type = None
        self._interpolation_code = None
        self._data_collection_type = None
        self._microphone_count = None
        self._directivity_angle = None
        self._perceived_noise_level = None
        self._tone_corrected_perceived_noise_level = None
        self._la = None
        self._lat = None
        self._effective_perceived_noise_level = None
        self._sound_exposure_level = None
        self._tone_corrected_sound_exposure_level = None
        self._tone_correction = None
        self._id = None
        self._operational_power_code = None
        self._operational_type_code = None
        self._engine_name = None
        self._engine_count = None
        self._drag_configuration = None
        self._source = None
        self._collection_date = None
        self._reference_distance = None
        self._reference_speed = None
        self._reference_temperature = None
        self._reference_humidity = None
        self._power_setting_description = None
        self._sound_pressure_levels = []
        self._frequency_resolution = 3
        self._minimum_frequency_band = 10
        self._maximum_frequency_band = 40

        if filename is not None:
            self.filename = filename
            f = open(self.filename)
            contents = f.readlines()
            f.close()
            self.load_data(contents, 0)

    @staticmethod
    def source_type():
        return 'FLIGHT'

    @property
    def frequency_resolution(self):
        return self._frequency_resolution

    @property
    def minimum_frequency_band(self):
        return self._minimum_frequency_band

    @property
    def maximum_frequency_band(self):
        return self._maximum_frequency_band

    @property
    def vehicle_type(self):
        return self._vehicle_type

    @property
    def interpolation_code(self):
        return self._interpolation_code

    @property
    def data_collection_type(self):
        return self._data_collection_type

    @property
    def number_microphone_in_average(self):
        return self._microphone_count

    @property
    def directivity_angle(self):
        return self._directivity_angle

    @property
    def perceived_noise_level(self):
        return self._perceived_noise_level

    @property
    def tone_corrected_pnl(self):
        return self._tone_corrected_perceived_noise_level

    @property
    def a_weighted_level(self):
        return self._la

    @property
    def tone_corrected_a_weighted_level(self):
        return self._lat

    @property
    def sound_exposure_level(self):
        return self._sound_exposure_level

    @property
    def tone_corrected_sel(self):
        return self._tone_corrected_sound_exposure_level

    @property
    def effective_perceived_noise_level(self):
        return self._effective_perceived_noise_level

    @property
    def tone_correction(self):
        return self._tone_correction

    @property
    def aircraft_id(self):
        return self._id

    @property
    def operational_power_code(self):
        return self._operational_power_code

    @property
    def operational_type_code(self):
        return self._operational_type_code

    @property
    def engine_name(self):
        return self._engine_name

    @property
    def engine_count(self):
        return self._engine_count

    @property
    def drag_configuration(self):
        return self._drag_configuration

    @property
    def source(self):
        return self._source

    @property
    def collection_date(self):
        return self._collection_date

    @property
    def reference_distance(self):
        return self._reference_distance

    @property
    def reference_speed(self):
        return self._reference_speed

    @property
    def reference_temperature(self):
        return self._reference_temperature

    @property
    def reference_humidity(self):
        return self._reference_humidity

    @property
    def power_setting_description(self):
        return self._power_setting_description

    @property
    def power_settings(self):
        return self._power_settings

    def calculate_sound_pressure_levels(self, band, elevation, polar):
        index = band - 10
        spls = np.zeros((len(elevation),))
        for i in range(0, len(elevation), 1):
            spls[i] = self._sound_pressure_levels[index]

        return spls

    def sound_pressure_level(self, location_or_elevation, polar=None):
        return self._sound_pressure_levels

    def write_data(self, file):
        if self.vehicle_type == AircraftType.Military:
            record1 = 'MILITARY FM{:5d}{:2d}{:1d}'.format(self.aircraft_id, self.operational_power_code,
                                                          self.operational_type_code)
        else:
            record1 = 'CIVILIAN FC{:5d}{:2d}{:1d}'.format(self.aircraft_id, self.operational_power_code,
                                                          self.operational_type_code)

        if self.interpolation_code == InterpolationCode.Fixed:
            record1 = record1 + 'F'
        elif self.interpolation_code == InterpolationCode.Parallel:
            record1 = record1 + 'P'
        elif self.interpolation_code == InterpolationCode.Variable:
            record1 = record1 + 'V'
        other1 = '{:<20}{:<20}{:>2} {:<25}'.format(self.name, self.engine_name, self.engine_count,
                                                   self.drag_configuration)
        if self.data_collection_type == DataCollectionType.Measured:
            other2 = 'MEASURED  '
        else:
            other2 = 'ESTIMATED '
        other1 = other1 + other2
        record2 = '{}{:<12}{:>11}{:>5} FT{:>5} KTS  {:>3} F  {:>3} PCT'.format(
            other1, self.source, self.collection_date, int(self.reference_distance.feet),
            int(self.reference_speed.knots), int(self.reference_temperature.fahrenheit),
            int(self.reference_humidity.percentage))
        record3 = '{:<20}'.format(self.power_setting_description)
        for setting in self.power_settings:
            record3 = record3 + '{:>9} {:<10} {:>9}{:>9}'.format('{:.2f}'.format(setting.value), setting.units,
                                                                 '{:.2f}'.format(setting.lower_limit),
                                                                 '{:.2f}'.format(setting.upper_limit))

        record4 = '{:>4}{:>6}{:>6}{:>6}{:>6}{:>6}{:>6}{:>6}{:>6}{:>6}'.format(self.number_microphone_in_average,
                                                                              self.directivity_angle.degrees,
                                                                              self.perceived_noise_level,
                                                                              self.tone_corrected_pnl,
                                                                              self.a_weighted_level,
                                                                              self.tone_corrected_a_weighted_level,
                                                                              self.effective_perceived_noise_level,
                                                                              self.sound_exposure_level,
                                                                              self.tone_corrected_sel,
                                                                              self.tone_correction)

        record5 = ''
        for level in self._sound_pressure_levels:
            record5 = record5 + '{:>4}'.format(int(level * 10))

        file.write(record1 + '\n')
        file.write(record2 + '\n')
        file.write(record3 + '\n')
        file.write(record4 + '\n')
        file.write(record5 + '\n')

    def load_data(self, contents, start_index):
        record1 = contents[start_index].rstrip()
        start_index += 1
        record2 = contents[start_index].rstrip()
        start_index += 1
        record3 = contents[start_index].rstrip()
        start_index += 1
        record4 = contents[start_index].rstrip()
        start_index += 1
        record5 = contents[start_index].rstrip()
        start_index += 1
        try:
            if record1[0:8] == 'MILITARY':
                self._vehicle_type = AircraftType.Military
            else:
                self._vehicle_type = AircraftType.Civilian
            self._id = int(record1[11:16])
            self._operational_power_code = int(record1[16:18])
            self._operational_type_code = int(record1[18:19])
            interp_code = record1[19:20]
            if interp_code == 'F':
                self._interpolation_code = InterpolationCode.Fixed
            elif interp_code == 'P':
                self._interpolation_code = InterpolationCode.Parallel
            elif interp_code == 'V':
                self._interpolation_code = InterpolationCode.Variable
            self.name = record2[0:20].strip()
            self._engine_name = record2[20:40].strip()
            self._engine_count = int(record2[40:42])
            self._drag_configuration = record2[43:43 + 25].strip()
            if record2[68:78].strip() == 'MEASURED':
                self._data_collection_type = DataCollectionType.Measured
            else:
                self._data_collection_type = DataCollectionType.Estimated
            self._source = record2[78:78 + 12].strip()
            self._collection_date = record2[90:90 + 11]
            self._reference_distance = Length(float(record2[101:101 + 5]), Length.Units.Feet)
            self._reference_speed = Speed(float(record2[109:109 + 5]), Speed.Units.Knots)
            self._reference_temperature = Temperature(float(record2[120:120 + 3]), Temperature.Units.Fahrenheit)
            self._reference_humidity = Humidity(float(record2[127:127 + 3]), Humidity.Units.Percentage)
            if len(record3) > 0:
                self._power_setting_description = record3[0:20].strip()
                record3 = record3[20:]
                while len(record3) > 0:
                    temp = NoiseFilePowerSetting()
                    temp.value = float(record3[0:9].strip())
                    temp.units = record3[10:10 + 10].strip()
                    temp.lower_limit = float(record3[20:20 + 9].strip())
                    temp.upper_limit = float(record3[30:30 + 9].strip())
                    self._power_settings.append(temp)
                    record3 = record3[39:]
            self._microphone_count = int(record4[0:4].strip())
            self._directivity_angle = Angle(float(record4[4:4 + 6].strip()))
            self._perceived_noise_level = float(record4[10:10 + 6].strip())
            self._tone_corrected_perceived_noise_level = float(record4[16:16 + 6].strip())
            self._la = float(record4[22:22 + 6].strip())
            self._lat = float(record4[28:28 + 6].strip())
            self._effective_perceived_noise_level = float(record4[34:34 + 6].strip())
            self._sound_exposure_level = float(record4[40:40 + 6].strip())
            self._tone_corrected_sound_exposure_level = float(record4[46:46 + 6].strip())
            self._tone_correction = float(record4[52:52 + 6].strip())

            for index in range(0, len(record5), 4):
                self._sound_pressure_levels.append(float(record5[index:index + 4].strip()) / 10.0)
        except Exception as e:
            raise ValueError('An error ocurred in the NOISEFILE formatted file at index = ' + str(
                start_index) + ' - error message: ' + str(e))
        return start_index


class NoiseFileStaticAcousticSource(NoiseFileAcousticSource):
    def __init__(self, filename=None):
        self.noise_suppression_system = ''
        self._reference_pressure = Pressure()

        if filename is None:
            super().__init__()
            self.polar_sound_pressure_levels = {}
        else:
            self.filename = filename
            self._power_settings = []
            self.polar_sound_pressure_levels = {}
            with open(filename, 'r') as file:
                contents = file.readlines()
                self._load_data(contents, 0)

    @property
    def measurement_count(self):
        return self.polar_sound_pressure_levels.shape[0]

    @property
    def reference_pressure(self):
        return self._reference_pressure

    def _load_data(self, contents, start_index):
        record1 = contents[start_index].rstrip()
        start_index += 1
        try:
            #   Record or line 1
            if record1[0:8] == 'MILITARY':
                self._vehicle_type = AircraftType.Military
            else:
                self._vehicle_type = AircraftType.Civilian
            self._id = int(record1[11:11 + 5])
            self._operational_power_code = int(record1[16:16 + 2])
            interpolation_code = record1[18:18 + 1]
            if interpolation_code == 'F':
                self._interpolation_code = InterpolationCode.Fixed
            elif interpolation_code == 'P':
                self._interpolation_code = InterpolationCode.Parallel
            elif interpolation_code == 'V':
                self._interpolation_code = InterpolationCode.Variable

            #   Record 2
            record2 = contents[start_index].rstrip()
            start_index += 1
            self.name = record2[0:20].strip()
            self._engine_name = record2[20:20 + 20].strip()
            self.noise_suppression_system = record2[40:40 + 14].strip()
            self._engine_count = int(record2[54:54 + 2])
            if 'MEASURED' in record2[57:57 + 10]:
                self._data_collection_type = DataCollectionType.Measured
            else:
                self._data_collection_type = DataCollectionType.Estimated
            self._source = record2[67:67 + 12].strip()
            self._collection_date = record2[79:79 + 11]

            #   Record 3
            record3 = contents[start_index].rstrip()
            start_index += 1
            self._power_setting_description = record3[0:20].strip()
            for i in range(0, 3, 1):
                substring = record3[20 + (i * 20): 20 + (i * 20) + 21]
                try:
                    power_setting_value = float(substring[0:9])
                    power_setting_units = substring[10:10 + 10].strip()
                    power_setting = NoiseFilePowerSetting()
                    power_setting.value = power_setting_value
                    power_setting.units = power_setting_units
                    self.power_settings.append(power_setting)
                except Exception as ee:
                    print('FAILURE: {}'.format(str(ee)))
                    break

            self._reference_distance = Length(float(record3[80:80 + 5]), Length.Units.Feet)
            self._reference_speed = Speed()
            self._reference_temperature = Temperature(float(record3[89:89 + 3]), Temperature.Units.Fahrenheit)
            self._reference_pressure = Pressure(float(record3[103:103 + 6]), Pressure.Units.InchesMercury)
            self._reference_humidity = Humidity(float(record3[95:95 + 3]))

            self.polar_sound_pressure_levels = np.zeros((19, 31))
            start_index += 1
            for band_index in range(10, 41, 1):
                data_record = contents[start_index].rstrip()
                start_index += 1
                temp_elements = data_record.split(' ')
                elements = [x for x in temp_elements if x != '']
                n = 1
                for angle_index in range(19):
                    self.polar_sound_pressure_levels[angle_index, band_index - 10] = float(elements[n]) / 10.0
                    n += 1
        except Exception as ee:
            raise ValueError('{} occurred in the NOISEFILE file at index = {}'.format(str(ee), start_index))
        return start_index

    def write_data(self, writer):
        if self.vehicle_type == AircraftType.Military:
            record = 'MILITARY FM{:5d}{:2d}'.format(self.aircraft_id, self.operational_power_code)
        else:
            record = 'CIVILIAN FC{:5d}{:2d}'.format(self.aircraft_id, self.operational_power_code)

        if self.interpolation_code == InterpolationCode.Fixed:
            record = record + 'F'
        elif self.interpolation_code == InterpolationCode.Parallel:
            record = record + 'P'
        elif self.interpolation_code == InterpolationCode.Variable:
            record = record + 'V'
        writer.write(record + '\n')

        record = '{:<20}{:<20}{:<14}{:>2} '.format(self.name, self.engine_name, self.noise_suppression_system,
                                                   self.engine_count)
        if self.data_collection_type == DataCollectionType.Measured:
            record = record + 'MEASURED  '
        else:
            record = record + 'ESTIMATED '

        record = record + '{:<12}{:>11} Single Engine Data'.format(self.source, self.collection_date)

        writer.write(record + '\n')
        record = '{:<20}'.format(self.power_setting_description)
        for setting in self.power_settings:
            record = record + '{:>9} {:<10}'.format('{:.2f}'.format(setting.value), setting.units)
        while len(record) < 80:
            record = record + ' '

        record = record + '{:>5} FT {:>3} F {:>3} PCT {:>6} IN HG'.format(int(self.reference_distance.feet),
                                                                          int(self.reference_temperature.fahrenheit),
                                                                          int(self.reference_humidity.percentage),
                                                                          self.reference_pressure.inches_mercury)
        writer.write(record + '\n')
        record = ' BAND  '
        for angle in range(0, 181, 10):
            record = record + '{:>5}'.format(angle)
        writer.write(record + '\n')

        for band_index in range(10, 41, 1):
            record = '  {:>2}   '.format(band_index)
            for angle_key in range(19):
                record = record + '{:>5}'.format(int(self.polar_sound_pressure_levels[angle_key, band_index - 10] * 10))

            writer.write(record + '\n')

    def sound_pressure_level(self, coord_or_polar, azimuthal=None) -> list:
        if isinstance(coord_or_polar, SphericalCoordinate):
            return self.__sound_pressure_level__(coord_or_polar.azimuthal)
        elif isinstance(coord_or_polar, Angle) and azimuthal is not None:
            return self.__sound_pressure_level__(azimuthal)
        else:
            raise ValueError("You must use the PythonCoordinates module to define the location on the source")

    def __sound_pressure_level__(self, polar) -> list:
        if isinstance(polar, Angle):
            angle = polar.normalized.degrees
            if angle > 180:
                angle = 360 - angle
            degree_angle_key = int(np.floor(angle / 10))
            return self.polar_sound_pressure_levels[degree_angle_key]
        else:
            raise ValueError("You must use the PythonCoordinates module to define the location on the source")


class FlightNoiseFileDatabase:
    def __init__(self, filename):
        """
        This constructor will load the entire noise database from the flight01.dat file that is passed to the 
        constructor
        :param filename: str - the path to the NOISEFILE database which we want to load 
        """
        self.filename = filename
        self.sources = []
        with open(filename, 'r') as f:
            contents = f.readlines()
        line_index = 0
        while line_index < len(contents):
            src_data = NoiseFileAcousticSource()
            try:
                line_index = src_data.load_data(contents, line_index)
            except Exception as eee:
                raise ValueError('An incorrect formatted NOISEFILE entry found at index {}: {}'.format(
                    line_index, str(eee)))
            self.sources.append(src_data)


class GroundNoiseFileDatabase:
    def __init__(self, path):
        """
        This will read a file with multiple NOISEFILE_Static_AcousticSource definition in a single file and place the
        elements into an internal array
        """

        self.filename = path
        self.sources = []

        with open(self.filename, 'rt') as f:
            contents = f.readlines()

        line_index = 0

        while line_index < len(contents):
            src_data = NoiseFileStaticAcousticSource()
            try:
                line_index = src_data.load_data(contents, line_index)
            except Exception as eee:
                raise ValueError("An incorrect formatted NOISEFILE entry found at index {}: {}".format(
                    line_index, str(eee)
                ))

            self.sources.append(src_data)



from PythonCoordinates.measurables.physical_quantities import Speed, Angle, Length
from PythonCoordinates.coordinates.coordinate_representations import CartesianCoordinate, SphericalCoordinate
from pytimbre.spectral.fractional_octave_band import FractionalOctaveBandTools as fob
import numpy as np


class IdealDipoleX:
    """
    This class implements the equation from Dr. Frank Mobley's dissertation that defines an acoustic dipole with an
    arbitrary rotation about the z-axis. It provides methods to alter the strength and frequency of the dipole so that
    a series of sound pressure levels can be obtained for any spherical angle.
    """

    def sound_pressure_level(self, location_or_elevation, polar: Angle = None):
        """
        Given a location on the surface of the directivity pattern, this will determine the sound pressure leve at each
        of the frequencies within the frequency list.

        Parameters
        ----------
        :param location_or_elevation:
            Angle or SphericalCoordinate or CartesianCoordinate - the first location of the point that we want to
            process
        :param polar: Angle
            if the first element is an angle, this defines the second SphericalCoordinate of the location that we
            want to process

        Returns
        -------
        :returns: np.ndarray - a collection of sound pressure levels for the location given
        """

        spl = np.zeros((len(self._frequency), ))

        if isinstance(location_or_elevation, Angle) and polar is None:
            raise ValueError("If the first argument is an angle, the second argument must also be an angle")
        elif isinstance(location_or_elevation, Angle) and isinstance(polar, Angle):
            coord = SphericalCoordinate(self.reference_distance, polar, location_or_elevation)
        elif isinstance(location_or_elevation, SphericalCoordinate):
            coord = location_or_elevation
        elif isinstance(location_or_elevation, CartesianCoordinate):
            coord = SphericalCoordinate(location_or_elevation)
        else:
            coord = SphericalCoordinate()

        for i in range(len(self._frequency)):
            wavelength = self._sound_speed.meters_per_second / self._frequency[i]
            k = 1 / wavelength
            w = 2 * np.pi * self._frequency[i]

            c_pressure = -complex(0, self._amplitude * k * self._separation / self.reference_distance.meters)
            c_pressure *= complex(np.cos(w * self._time - k * self.reference_distance.meters),
                                  np.sin(w * self._time - k * self.reference_distance.meters))
            c_pressure *= np.cos(coord.polar) * np.sin(coord.azimuthal)
            pressure = abs(c_pressure)

            spl[i] = 20 * np.log10(pressure / 20e-6)

        return spl

    def __init__(self, dipole_separation: float = 0.001, amplitude: float = 0.1,
                 sound_speed: Speed = Speed.ref_speed_of_sound(), frequency: np.ndarray = fob.tob_frequencies(),
                 ref_distance: Length = Length.reference_source_radius()):
        """
        This constructor defines the initial conditions for the dipole class

        Parameters
        ----------

        :param dipole_separation: float
            the distance between the two monopoles that form the dipole
        :param amplitude: float
            the strength of both monopoles
        :param sound_speed: Speed
            This is the nominal speed of sound for the determination of the wave number
        """

        self._separation = dipole_separation
        self._amplitude = amplitude
        self._sound_speed = sound_speed
        self._frequency = frequency
        self.reference_distance = ref_distance
        self.name = "Dr. Mobley's Ideal Dipole (X-axis orientation)"
        self.minimum_frequency_band = self._frequency[0]
        self.maximum_frequency_band = self._frequency[-1]
        self._time = 0

    @property
    def calculation_time(self):
        return self._time

    @calculation_time.setter
    def calculation_time(self, value):
        self._time = value




import matplotlib.pyplot as plt
import pandas as pd
import xml.etree.ElementTree
import os.path
from PythonCoordinates.measurables.physical_quantities import Length, Angle, Temperature, Pressure, Humidity
from PythonCoordinates.measurables.Measurable import InvalidUnitOfMeasureException
from PythonCoordinates.coordinates.coordinate_representations import CartesianCoordinate, SphericalCoordinate
from pytimbre.spectral.fractional_octave_band import FractionalOctaveBandTools
import numpy as np
from dateutil import parser
import warnings
import sys
import scipy
from matplotlib import cm, colors


def plot_irregular_surface(locations, desired_x=None, desired_y=None, ax=None, vmin=None, vmax=None, colormap=cm.jet,
                           invert_z_axis: bool = True):
    """
    We desire the ability to plot the spherical surface in a 3-D representation where the faces will be the color of
    the acoustic level that we are attempting to characterize.
    :param desired_y:
    :param desired_x:
    :param invert_z_axis: bool
        This flag determines whether the z-axis values need to be negated for the proper display of the data
    :param colormap:
    :param vmax: float
        The maximum value for the coloring of the faces of the surface
    :param vmin: float
        The minimum value for the coloring of the faces of the surface
    :param locations: list or array-like
        Collection of spherical locations that we want to display one a spherical surface.
    :param ax: Matplotlib.pyplot.axis
        The axis that we want to use for the plotting
    :return:
        axis object that hold the data, and the smoothed surface
    """
    from PythonCoordinates.conversions.measurement_resolution import nearest_neighbor_dense_sampling

    if ax is None:
        fig, ax = plt.subplots(subplot_kw={"projection": "3d"})

    #   Use the nearest neighbor interpolation to define the surface that we want to plot.
    #
    #   Examine the data and determine the range of the azimuthal and polar angles within the data
    true_data = _transform_data_to_dataframe(locations)

    if desired_x is None:
        min_az = true_data['x'].min()
        max_az = true_data['x'].max()
        desired_azimuthal = np.linspace(min_az, max_az, 50)
    else:
        desired_azimuthal = desired_x

    if desired_y is None:
        min_polar = true_data['y'].min()
        max_polar = true_data['y'].max()
        desired_polar = np.linspace(min_polar, max_polar, 50)
    else:
        desired_polar = desired_y

    #   Build the ability to interpolate the colors
    if vmax is None:
        vmax = true_data['z'].min()
    if vmin is None:
        vmin = true_data['z'].max()

    color_normalizer = colors.Normalize(vmin=vmin, vmax=vmax)

    #   Calculate the interpolated grid in spherical coordinates
    xx, yy, szz, zz = nearest_neighbor_dense_sampling(true_data, desired_azimuthal, desired_polar,
                                                      smoothing_error_tolerance=1e-4, verbose_info=False)

    #   Now we need to create a set of locations that are in Cartesian coordinate representation, but still have the
    #   same shape as the spherical data
    cxx, cyy, czz = _transform_data_to_matrix(xx, yy, szz)

    if invert_z_axis:
        czz *= -1

    ax.plot_surface(cxx, cyy, czz, facecolors=colormap(color_normalizer(szz)))

    return ax, xx, yy, szz, color_normalizer


def _transform_data_to_dataframe(locations: np.ndarray):
    """
    This function assists in reducing the complexity of the plotting functions and transforms the SphericalCoordinate
    array to a Pandas.DataFrame that contains the information in a flat plane to be interpolated by the nearest
    neighbor interpolation routine.
    :param locations: Numpy.ndarray
        This is the collection of SphericalCoordinate classes that needs to be converted to a Pandas.DataFrame. The
        values will be represented in the DataFrame as Cartesian coordinates, but this is in name only.
    :return: Pandas.DataFrame
    """

    true_data = pd.DataFrame(columns=['x', 'y', 'z'], index=np.arange(len(locations)))

    for i in range(len(locations)):
        if not isinstance(locations[i], SphericalCoordinate):
            raise ValueError("The surface locations must be a SphericalCoordinate")

        true_data.iloc[i, :] = [locations[i].azimuthal.degrees, locations[i].polar.degrees, locations[i].r.meters]

    return true_data


def _transform_data_to_matrix(xx, yy, zz):
    """
    Data is returned from the nearest neighbor interpolation as doubly dimensioned matrices. However,
    this information will be stored in whatever format the data was presented in. For the plotting of the surfaces
    the information is obtained as a series of angles and lengths that represent a Spherical Coordinate
    representation. This function will transform the spherical representation to a doubly dimensioned matrix using a
    CartesianCoordiante representation
    :param xx: Numpy.ndarray
        The decimal degrees values for the azimuthal direction
    :param yy: Numpy.ndarray
        The decimal degrees values for the polar direction
    :param zz: Numpy.ndarray
        The decimal meters values for the radial direction
    :return:
        Cartesian xx, yy, zz matrices in units of meters
    """
    cxx = np.zeros(xx.shape)
    cyy = np.zeros(yy.shape)
    czz = np.zeros(zz.shape)

    for i in range(xx.shape[0]):
        for j in range(xx.shape[1]):
            c = SphericalCoordinate(r=Length(zz[i, j]),
                                    polar=Angle(yy[i, j]),
                                    azimuthal=Angle(xx[i, j])).to_cartesian()
            cxx[i, j] = c.x.meters
            cyy[i, j] = c.y.meters
            czz[i, j] = c.z.meters

    return cxx, cyy, czz


def plot_unwrapped(xx, yy, zz, color_map_name: str = 'jet', contour_levels: np.ndarray = np.arange(0, 100, 10),
                   contour_lines: np.ndarray = np.arange(50, 100, 5)):
    """
    This function will plot the data as an unwrapped display of the spherical data.
    :param xx: Numpy.ndarray
        The x-values to plot
    :param yy: Numpy.ndarray
        The x-values to plot
    :param zz: Numpy.ndarray
        The surface-values to plot
    :param color_map_name: str
        The name of the colormap to use for the surface
    :param contour_levels: Numpy.ndarray
        A linear array of values that we will use to color the faces of the surface
    :param contour_lines: Numpy.ndarray
        A collection of value to calculate contour lines at
    :return:
        the figure
    """

    #   Create the figure
    fig, ax = plt.subplots()

    #   Build the filled contour plot
    contours = ax.contourf(xx,
                           yy,
                           zz,
                           levels=contour_levels,
                           cmap=color_map_name,
                           antialiased=True)
    cbar = plt.colorbar(contours)

    #   Add the contour lines
    cs = ax.contour(xx,
                    yy,
                    zz,
                    levels=contour_lines,
                    linestyles='solid',
                    colors='k')
    ax.clabel(cs, fontsize=6, inline=1, fmt='%.0f')

    return fig, ax


class HarmonicSeries:
    """
    The harmonic source is a collection of coefficient definition sets. This definition is the example of how to
    organize the information, but does not have to be a frequency band. So this was refactored to be more generic.
    """

    def __init__(self, series_coefficients: np.ndarray, description: str, r: float = None, mse: float = None):
        """
        This constructor builds the object for the holding of information regarding the construction of a frequency's
        acoustic surface
        :param mse: float
            The mean squared error determined during the curve fitting
        :param r: float
            The pearson correlation coefficient determined during the curve fitting
        :param description: str
            a text description of the object represented by this series expansion
        :param series_coefficients: complex, array-like
            the list of harmonic coefficients for the reconstruction
        """

        self._coefficients = series_coefficients
        self._description = description
        self._order = int(np.floor(np.sqrt(len(self._coefficients)) - 1))
        self._pearson_r = r
        self._mse = mse

    @property
    def coefficients(self):
        """
        If the user needs to explore the coefficients directly, they can use this property to gain access to the
        array-like list of complex coefficients.
        :return: array-like
            List of complex coefficients that represent the harmonic expansion
        """
        return self._coefficients

    @property
    def series_description(self):
        """
        Initially, these series represented a frequency within the sound pressure level spectrum. However,
        there was no reason to require this representation has a sound pressure level representation. Rather it is
        merely an expansion of a spherical dataset. As such this could be any function. So the band number was
        dropped and a string description was used to determine the function expanded with the harmonic series.
        :return: str
            The custom description of the series expansion
        """
        return self._description

    @series_description.setter
    def series_description(self, value):
        self._description = value

    @property
    def fit_correlation_coefficient(self):
        return self._pearson_r

    @property
    def fit_mse(self):
        return self._mse

    @property
    def order(self):
        """
        The order refers to the harmonic series expansion order that is represented by the coefficients within this
        class. While this can be easily calculated from the length of the coefficient array, it was determined in the
        constructor so that it is quicker to determine the limits of the expansion.
        :return: int
            The order ("l") within the harmonic series expansion.
        """
        return self._order

    @staticmethod
    def parse_xml(root):
        """
        This function will parse through the information in the XML node and return a HarmonicSeries object to the user.
        :param root: ElementTree
        :return: HarmonicSeries object
        """

        #   Loop through the coefficient list
        coefficient_list = list()
        for coefficients in root:
            for coefficient in coefficients:
                coefficient_list.append(complex(float(coefficient[0].text), float(coefficient[1].text)))

        if "description" in root.attrib.keys():
            description = root.attrib['description']
        else:
            description = None

        if "fit_mse" in root.attrib.keys():
            mse = root.attrib['fit_mse']
        else:
            mse = None

        if "fit_correlation_coefficient" in root.attrib.keys():
            fit_r = root.attrib["fit_correlation_coefficient"]
        else:
            fit_r = None

        series = HarmonicSeries(series_coefficients=np.asarray(coefficient_list), description=root.attrib[
            'description'], r=fit_r, mse=mse)

        return series

    @staticmethod
    def yml(l, m, polar, azimuthal):
        """
        Calculate the normalized spherical harmonics for the provided order (l) and power (m) at the provided angle set.
        The original function was replaced with the scipy function for the calculation of the spherical harmonics.

        :param l : int
            the order of the series
        :param m : int
            the power of the series
        :param polar : double (units: radians)
            the polar angle of the spherical harmonics
        :param azimuthal : double (units: radians)
            the azimuthal angle of the spherical harmonics

        :returns : complex
            the values of the spherical harmonics at these angles, order and power.

        20220329 - FSM - Refactored the calculation of the angles that are used within the determination of the
            spherical harmonic value
        """
        if isinstance(azimuthal, Angle):
            az_angle_radians = azimuthal.radians
        elif isinstance(azimuthal, tuple):
            if isinstance(azimuthal[0], Angle):
                az_angle_radians = azimuthal[0].radians
            else:
                az_angle_radians = azimuthal[0]
        else:
            az_angle_radians = azimuthal

        if isinstance(polar, Angle):
            pol_angle_radians = polar.radians
        elif isinstance(polar, tuple):
            if isinstance(polar[0], Angle):
                pol_angle_radians = polar[0].radians
            else:
                pol_angle_radians = polar[0]
        else:
            pol_angle_radians = polar

        result = scipy.special.sph_harm(m, l, az_angle_radians, pol_angle_radians)

        return result

    @staticmethod
    def harmonic_matrix(order, polar, azimuthal):
        """
        Using the matrix format, create a matrix that has a row for each element in the theta and phi vectors, and a
        column for each of the l, m pairs up to the (l**2 + 1) count.

        order : int
            The maximum order of the series expansion and the definition of the elements to calculate
        polar : Angle or double, array-like
            the polar angle of the series expansion
        azimuthal : Angle or double, array-like
            the azimuthal angle of the series expansion

        returns: complex, array-like
            the matrix that is nxm, where n = len(polar) and m = (l**2+1).
        """

        #   If both the polar and azimuthal arguments are of type Angle

        if isinstance(polar, Angle) and isinstance(azimuthal, Angle):
            #   Create the return object
            b = HarmonicSeries._build_even_order_matrix(order, [polar], [azimuthal])

            #   Using the positive values of the power (m), look for the symmetric values to determine the negative
            #   values of the power (m)
            b = HarmonicSeries._build_reflected_matrix(b, order, [polar])

        else:
            #   Create the return object
            b = HarmonicSeries._build_even_order_matrix(order, polar, azimuthal)

            #   Using the positive values of the power (m), look for the symmetric values to determine the negative
            #   values of the power (m)
            b = HarmonicSeries._build_reflected_matrix(b, order, polar)

        return b

    @staticmethod
    def _build_even_order_matrix(order: int, polar, azimuthal):
        """
        This function will build the positive 'm' values within the harmonic series matrix. This is because the
        calculation of the spherical harmonics is non-trivial and exponentially increasing with time. However,
        there is a relationship between the +m and -m values where, knowing the value of the +m harmonic, the -m harmonic
        is easily calculated with symmetry and conjugation.
        :param order: int
            The order of the harmonic - this is the L index for which we calculate the m values on the interval 0 <= m <= l
        :param polar: list or Numpy.ndarray
            The collection of polar angles that we want to calculate the harmonics at
        :param azimuthal: list or Numpy.ndarray
            The collection of azimuthal angles that we want to calculate the harmonics at
        :return: np.ndarray
            The dense matrix of values for positive values of m
        """

        b = np.zeros(shape=(len(polar), ((order + 1) * (order + 1))), dtype=np.complex_)

        #   Calculate the values of the spherical harmonics with a positive power (m)

        for index_element in range(len(polar)):
            for index_order in range(order + 1):
                for index_m in range(index_order + 1):
                    #   Calculate the index within the matrix for this l, m combination
                    index = index_order * index_order + index_order + index_m

                    #   Insert the harmonic value at this site
                    b[index_element, index] = HarmonicSeries.yml(
                        index_order,
                        index_m,
                        polar[index_element],
                        azimuthal[index_element])

        return b

    @staticmethod
    def _build_reflected_matrix(b: np.ndarray, order: int, polar):
        """
        This function reflects the upper triangle to the lower triangle of the matrix of harmonic values. This is
        accomplished due to the relationship of the m and -m values of the harmonic series.
        :param b: numpy.ndarray
            This is the matrix that have been calculated for the positive mode values (m)
        :param order: int
            The number of harmonics to use (l)
        :param polar:
            The collection of polar angles. This is only used to provide a limit for the integration and summation
        :return:
            The fully built matrix is returned.
        """
        for index_element in range(len(polar)):
            for index_order in range(order + 1):
                for index_m in range(-index_order, 0, 1):
                    #   Calculate the index of this element, and the symmetric +m value
                    index = index_order * index_order + index_order + index_m
                    sym_index = index_order * index_order + index_order + abs(index_m)

                    #   get the symmetric value and adjust the complex value
                    b[index_element, index] = np.power(-1.0, index_m) * complex.conjugate(
                        b[index_element, sym_index])

        return b

    @staticmethod
    def invertible_harmonic_matrix(order, polar, azimuthal):
        """
        Inherently the matrix B is non-square.  To invert this matrix we must generate a square matrix.  For this we
        use the complex conjugate transpose of B and multiply by B, i.e. BtB.

        order : int
            The maximum order of the series expansion and the definition of the elements to calculate
        polar : Angle or double, array-like
            the polar angle of the series expansion
        azimuthal : Angle or double, array-like
            the azimuthal angle of the series expansion

        returns: complex, array-like
            the matrix that is mxm, where m = (l**2+1).
        """

        #   Get the matrix
        b = np.asmatrix(HarmonicSeries.harmonic_matrix(order, polar, azimuthal))

        #   Determine the complex conjugate transpose
        trans = b.getH()

        #   Multiply with the harmonic matrix to obtain the BtB square matrix
        mul = trans.dot(b)

        return np.asarray(mul)

    @staticmethod
    def scaled_surface_value(b, r):
        """
        The R.H.S. of the matrix equation evaluates to BtR and this will need to be computed from the harmonic matrix
        and the radial sound pressure level values.

        b : complex, array-like
            the harmonic series matrix that is nxm, where n is the number of elements on the surface of the sphere, and
            m is the number of harmonic components (m = (order + 1)**2)

        r : double, array-like
            the sound pressure level radial function at a selected frequency

        returns : complex, array-like
            the complex conjugate transpose of the harmonic matrix multiplied by the radial values (i.e. BtR)
        """
        if (isinstance(b, np.ndarray) or isinstance(b, np.matrix)) and (
                isinstance(r, np.ndarray) or isinstance(r, list)):
            bt = np.asmatrix(b).getH()
            btr = np.matmul(bt, r)

            return np.asarray(btr).reshape((b.shape[1],))

    def predict(self, location):
        """
        This function will calculate the value of the harmonic series expansion at the specified location.
        :param location: SphericalCoordinate
            The location on the surface of the sphere where we want to determine the value of the harmonic series.
            Since the data is expanded as a sphere, we will require the user to represent the location as a
            SphericalCoordinate object
        :return: float
            The value of the harmonic series at the specific location provided
        """

        #   Determine the harmonic matrix at this location
        if isinstance(location, SphericalCoordinate):
            b = HarmonicSeries.harmonic_matrix(self.order, location.polar, location.azimuthal)
        elif (isinstance(location, np.ndarray) or isinstance(location, list)) and \
                isinstance(location[0], SphericalCoordinate):
            polar = np.empty(len(location), dtype=Angle)
            azimuthal = np.empty(len(location), dtype=Angle)
            for i in range(len(location)):
                polar[i] = location[i].polar
                azimuthal[i] = location[i].azimuthal

            b = HarmonicSeries.harmonic_matrix(self.order, polar, azimuthal)

        return b.dot(self.coefficients).real

    def make_xml(self, root):
        """
        In effort to utilize the proper construction of the XML nodes, we will provide the node that represents this
        class's series expansion and append it directly to the root that is provided.
        :param root: ElementTree
            The parent of the XML node that we are constructing in this function
        :return:
        """

        parent = xml.etree.ElementTree.SubElement(root, "series_expansion")
        parent.attrib['description'] = self.series_description
        if self.fit_correlation_coefficient is not None:
            parent.attrib['fit_correlation_coefficient'] = "{:.10f}".format(self.fit_correlation_coefficient)
        if self.fit_mse is not None:
            parent.attrib['fit_mse'] = "{:.10f}".format(self.fit_mse)

        coefficients = xml.etree.ElementTree.SubElement(parent, "coefficients")

        for c in self.coefficients:
            node = xml.etree.ElementTree.SubElement(coefficients, "coefficient")

            real_node = xml.etree.ElementTree.SubElement(node, "real")
            real_node.text = str(c.real)

            imag_node = xml.etree.ElementTree.SubElement(node, "real")
            imag_node.text = str(c.imag)

    @staticmethod
    def fit(x, y, order: int):
        """
        This function will convert the
        :param x: Numpy.ndarray of SphericalCoordinate objects
            This is a collection of the known locations on the spherical surface. At this point, the radius is ignored
        :param y: Numpy.ndarray
            A single or double dimensioned array that will be used to define the radial value of the harmonic series
            expansion.
        :param order: int
            The maximum order of fitting that will be accomplished
        :return: HarmonicSeries
            The collection of coefficients that best fit this array of location/radial values
        """
        from scipy.stats import pearsonr
        from sklearn.metrics import mean_squared_error

        if isinstance(order, int):
            o = order
        else:
            o = int(np.floor(order))

        if not (isinstance(x, np.ndarray) or isinstance(x, list)) and not isinstance(x[0], SphericalCoordinate):
            raise ValueError("The x values must be a list or numpy array of SphericalCoordinate objects")

        if not (isinstance(y, np.ndarray) or isinstance(y, list)):
            raise ValueError("The y values must be a numpy array")

        #   Extract the polar and azimuthal angles from the collection
        if isinstance(x, list):
            length = len(x)
        else:
            length = x.shape[0]
        polar = np.empty(length, dtype=Angle)
        azimuthal = np.empty(length, dtype=Angle)
        for i in range(length):
            polar[i] = x[i].polar
            azimuthal[i] = x[i].azimuthal

        #   Calculate the matrix solution to the fitting
        b = HarmonicSeries.harmonic_matrix(o, polar, azimuthal)
        btb = HarmonicSeries.invertible_harmonic_matrix(o, polar, azimuthal)
        inv_btb = np.linalg.inv(btb + sys.float_info.epsilon)
        btf = HarmonicSeries.scaled_surface_value(b, y)

        c = inv_btb.dot(btf)
        lvl = b.dot(c).real
        r = pearsonr(y, lvl)[0]
        mse = mean_squared_error(y, lvl)
        return HarmonicSeries(c, "", r, mse)


class HarmonicSpectrumNoiseSource:
    """
    In effort to create a class that contains the information for a sound pressure level spectrum, this class
    contains a collection of HarmonicSeries objects that represent each frequency in the spectrum. In the future,
    additional classes can be derived that will represent different methods or combinations of functions on the
    surface of a sphere.
    """

    def __init__(self, a=None):
        """
        This constructor will build a dataset from the file that is passed as an argument to the constructor, or build
        the blank contents of the class for the construction of the directivity pattern and the file later.
        :param a: str - the path to the file that is to be read
        """

        self._de_dopplerized = False
        self._creation_date = None
        self._data_measured = 1
        self._measure_date = None
        self._version = "XML v1.1.0.1"
        self._power_setting = None
        self._power_setting_units = None

        self._reference_temperature = Temperature.std_temperature()
        self._reference_pressure = Pressure.std_pressure()
        self._reference_humidity = Humidity.std_humidity()
        self._reference_distance = Length.reference_sphere_radius()
        self._frequency_resolution = 3
        self._minimum_frequency_band = 10
        self._maximum_frequency_band = 40
        self._coefficients = []

        self._filename = None

        if a is not None:
            if not (isinstance(a, str) or isinstance(a, xml.etree.ElementTree.Element)):
                raise ValueError()
            else:
                if isinstance(a, str):
                    if not os.path.exists(a):
                        raise FileNotFoundError()

                    self._filename = a
                    self._load_definition(a)
                elif isinstance(a, xml.etree.ElementTree.Element):
                    self._load_definition(a)

                self._build_reference_atmospheric_absorption_losses()

    @property
    def is_de_dopplerized(self):
        return self._de_dopplerized

    @property
    def creation_date(self):
        return self._creation_date

    @creation_date.setter
    def creation_date(self, value):
        import dateutil.parser

        if isinstance(value, str):
            try:
                value = dateutil.parser.parse(value)
            except dateutil.parser.ParserError:
                return

        self._creation_date = value

    @property
    def is_measured(self):
        return self._data_measured

    @is_measured.setter
    def is_measured(self, value):
        self._data_measured = int(value)

    @property
    def measurement_date(self):
        return self._measure_date

    @measurement_date.setter
    def measurement_date(self, value):
        import dateutil.parser

        if isinstance(value, str):
            try:
                value = dateutil.parser.parse(value)
            except dateutil.parser.ParserError:
                return

        self._measure_date = value

        self._measure_date = value

    @property
    def version(self):
        return self._version

    @version.setter
    def version(self, value):
        self._version = value

    @property
    def aircraft_power_setting(self):
        return self._power_setting

    @aircraft_power_setting.setter
    def aircraft_power_setting(self, value):
        if isinstance(value, str):
            try:
                value = float(value)
            except ValueError:
                return

        self._power_setting = value

    @property
    def aircraft_power_setting_units(self):
        return self._power_setting_units

    @aircraft_power_setting_units.setter
    def aircraft_power_setting_units(self, value):
        self._power_setting_units = value

    @property
    def ref_measurement_distance(self):
        return self._reference_distance

    @ref_measurement_distance.setter
    def ref_measurement_distance(self, value):
        if isinstance(value, str):
            value = Length(float(value), Length.Units.Meters)
        elif isinstance(value, float):
            value = Length(value, Length.Units.Meters)

        self._reference_distance = value

    @property
    def ref_humidity(self):
        return self._reference_humidity

    @ref_humidity.setter
    def ref_humidity(self, value):
        if isinstance(value, str):
            value = Humidity(float(value))
        elif isinstance(value, float):
            value = Humidity(value, Humidity.Units.Percentage)

        self._reference_humidity = value

    @property
    def ref_pressure(self):
        return self._reference_pressure

    @ref_pressure.setter
    def ref_pressure(self, value):
        if isinstance(value, str):
            value = Pressure(float(value))
        elif isinstance(value, float):
            value = Pressure(value, Pressure.Units.InchesMercury)

        self._reference_pressure = value

    @property
    def ref_temperature(self):
        return self._reference_temperature

    @ref_temperature.setter
    def ref_temperature(self, value):
        if isinstance(value, float):
            value = Temperature(value, Temperature.Units.Kelvin)
        elif isinstance(value, str):
            value = Temperature(float(value), Temperature.Units.Kelvin)

        self._reference_temperature = value

    @property
    def fractional_resolution(self):
        return self._frequency_resolution

    @fractional_resolution.setter
    def fractional_resolution(self, value):
        if isinstance(value, str):
            value = int(value)
        self._frequency_resolution = value

    @property
    def min_freq_band(self):
        return self._minimum_frequency_band

    @min_freq_band.setter
    def min_freq_band(self, value):
        if isinstance(value, str):
            value = int(value)

        self._minimum_frequency_band = value

    @property
    def max_freq_band(self):
        return self._maximum_frequency_band

    @max_freq_band.setter
    def max_freq_band(self, value):
        if isinstance(value, str):
            value = int(value)

        self._maximum_frequency_band = value

    @property
    def orders(self):
        orders = list()

        for band in self._coefficients:
            orders.append(band.order)

        return orders

    @property
    def name(self):
        return os.path.splitext(self.filename)[0]

    @property
    def minimum_frequency(self):
        return FractionalOctaveBandTools.center_frequency(self._frequency_resolution, self._minimum_frequency_band)

    @property
    def maximum_frequency(self):
        return FractionalOctaveBandTools.center_frequency(self._frequency_resolution, self._maximum_frequency_band)

    @property
    def series_coefficients(self):
        return self._coefficients

    @series_coefficients.setter
    def series_coefficients(self, value):
        self._coefficients = value

    @property
    def filename(self):
        return self._filename

    @property
    def reference_atmospheric_losses(self):
        return self._reference_loss

    def _load_definition(self, a):
        """
        This will load the definition from the XML file
        :param a: str
            the path to the definition file

        20220330 - FSM - Added some error catching for when attributes are not actually specified
        """
        attribute_names = ['frequency_resolution',
                           'creation_date',
                           'data_measured',
                           'measure_date',
                           'version',
                           'power_setting',
                           'power_setting_units',
                           'reference_temperature',
                           'reference_pressure',
                           'reference_humidity',
                           'reference_distance',
                           'minimum_frequency_band',
                           'maximum_frequency_band']
        property_names = ['fractional_resolution',
                          'creation_date',
                          'is_measured',
                          'measurement_date',
                          'version',
                          'aircraft_power_setting',
                          'aircraft_power_setting_units',
                          'ref_temperature',
                          'ref_pressure',
                          'ref_humidity',
                          'ref_measurement_distance',
                          'min_freq_band',
                          'max_freq_band']

        if isinstance(a, str):
            tree = xml.etree.ElementTree.ElementTree()
            tree.parse(a)
            root = tree.getroot()
        elif isinstance(a, xml.etree.ElementTree.Element):
            root = a

        #   We have constructed two arrays that contain a 1-to-1 relationship between the name of the attribute in
        #   the XML structure to the name of the property. We can now use the getattr and setattr functions to put
        #   the data in the correct place.
        for name in root.attrib.keys():
            element_index = attribute_names.index(name)

            setattr(self, property_names[element_index], root.attrib[name])

        #   Now the only remaining node is the coefficients node
        self._read_coefficients(root[0])

    def _read_coefficients(self, root):
        """
        This function will read the series_coefficients nodes from the file and create the contents for the
        coefficients of this source description
        :param root: xml.etree.ElementTree.Element
            This is the container for the series_expansion nodes that define the approximation for the various
            frequencies
        """
        for coefficient_storage in root:
            self._coefficients.append(HarmonicSeries.parse_xml(coefficient_storage))

    def _build_reference_atmospheric_absorption_losses(self):
        """
        Calculate the acoustic losses for the atmosphere at the reference conditions
        """
        from physical_propagation.acoustic_losses import AtmosphericAbsorption
        from pytimbre.spectral.fractional_octave_band import FractionalOctaveBandTools as fob

        self._reference_loss = list()
        for i in range(0, len(self._coefficients)):
            f = fob.center_frequency(self.fractional_resolution, i + self.min_freq_band)
            self._reference_loss.append(AtmosphericAbsorption.alpha(self.ref_temperature,
                                                                    self.ref_pressure,
                                                                    self.ref_humidity,
                                                                    f) * self.ref_measurement_distance.meters)

    def make_xml(self):
        """
        This will create the XML structure from the data within the class.
        :return: xml.etree.ElementTree.Element
        """
        #   Define the attribute and property names so that we can loop through these elements easily
        attribute_names = ['frequency_resolution',
                           'creation_date',
                           'data_measured',
                           'measure_date',
                           'version',
                           'power_setting',
                           'power_setting_units',
                           'reference_temperature',
                           'reference_pressure',
                           'reference_humidity',
                           'reference_distance',
                           'minimum_frequency_band',
                           'maximum_frequency_band']
        property_names = ['fractional_resolution',
                          'creation_date',
                          'is_measured',
                          'measurement_date',
                          'version',
                          'aircraft_power_setting',
                          'aircraft_power_setting_units',
                          'ref_temperature',
                          'ref_pressure',
                          'ref_humidity',
                          'ref_measurement_distance',
                          'min_freq_band',
                          'max_freq_band']

        #   Create the tree and the root element
        root = xml.etree.ElementTree.Element("sound_pressure_level_harmonic_series_definition")

        #   Loop through the attributes/properties and add them to the root node
        for i in range(len(attribute_names)):
            value = getattr(self, property_names[i])

            if not isinstance(value, str):
                if isinstance(value, Length):
                    value = str(value.meters)
                elif isinstance(value, Temperature):
                    value = str(value.kelvin)
                elif isinstance(value, Pressure):
                    value = str(value.kilopascal)
                elif isinstance(value, Humidity):
                    value = str(value.percentage)
                else:
                    value = str(value)

            root.attrib[attribute_names[i]] = value

        #   Now create the node that will hold the coefficient definitions.
        series_collection = xml.etree.ElementTree.SubElement(root, "frequency_definitions")

        #   Loop through the series collection within the class and build the nodes. Then add the nodes to this node
        for coefficients in self.series_coefficients:
            coefficients.make_xml(series_collection)

        return root

    def save(self, a: str = None):
        """
        Write the data to an XML file
        :param a: str - the path to the output file
        """

        tree = xml.etree.ElementTree.ElementTree(self.make_xml())
        tree.write(file_or_filename=a, xml_declaration=True, encoding="utf-8")

    def predict(self, location):
        """
        This function calculates the sound pressure level at a specific location and returns the entire spectrum
        :param location: SphericalCoordinate
            it is either the specific location on the surface as defined in Cartesian or spherical coordinates or the
            azimuthal angle
        :return: double, array-like
            the sound pressure level @ all frequencies within the definition
        """

        #   We expect that the user will want the complete set of information at the specific location that is
        #   provided as an argument. So we will create an array of values for each element in the collection of
        #   series coefficients.
        if isinstance(location, SphericalCoordinate):
            spl = np.zeros((len(self.series_coefficients), 1))
        else:
            spl = np.zeros((len(self.series_coefficients), len(location)))

        #   Loop through the series approximations and create the list of values that each creates.
        for i in range(len(self.series_coefficients)):
            spl[i, :] = self.series_coefficients[i].predict(location)

        return spl.transpose()

    @staticmethod
    def fit(x, y, maximum_order: int = 9):
        """
        This function expects the X values to be a single dimension array of SphericalCoordinate locations that will
        be transformed into a collection of angles. The y values must be a collection of 1-D data the same length as
        the x values. If y has a second dimension, then it will be iterated through and a number of series expansions
        will be created
        :param x: Numpy.ndarray of SphericalCoordinate objects
            This is a collection of the known locations on the spherical surface. At this point, the radius is ignored
        :param y: Numpy.ndarray
            A single or double dimensioned array that will be used to define the radial value of the harmonic series
            expansion.
        :param maximum_order: int
            The maximum order of fitting that will be accomplished
        """

        series = HarmonicSpectrumNoiseSource()

        #   ensure that the inputs are correct
        if not isinstance(x, np.ndarray):
            x = np.asarray(x)
        if not isinstance(x, np.ndarray) or not isinstance(x[0], SphericalCoordinate):
            raise ValueError("You must provide an array of SphericalCoordinate objects for the x argument")

        if x.shape[0] != y.shape[0]:
            raise ValueError("The first dimension of the x and y arrays must be the same!")

        #   Now loop through the second dimension of the y values
        if len(y.shape) == 1:
            series.series_coefficients.append(HarmonicSeries.fit(x, y, maximum_order))
        else:
            for i in range(y.shape[1]):
                series.series_coefficients.append(HarmonicSeries.fit(x, y[:, i], maximum_order))

        return series



from setuptools import setup, find_packages

with open('README.md') as readme_file:
    README = readme_file.read()

    with open('HISTORY.md') as history_file:
        HISTORY = history_file.read()

setup(
    name='physical_sources',
    version='0.7.3',
    packages=find_packages('src', exclude=['test*.py']),
    url='',
    license='',
    author='Dr. Frank Mobley',
    author_email='frank.mobley.1@afrl.af.mil',
    description="A collection of classes that can be used to build acoustic sources from the NOISEFILE format. It "
                "also contains classes to read the binary representations born from the author's dissertation",
    package_dir={'': 'src'},
    long_description=README + '\n\n' + HISTORY,
    long_description_content_type="text/markdown",
    install_requires=['numpy', 'scipy', 'PythonCoordinates>=0.5.0', 'Pytimbre>0.6.1', 'mne>=1.0.0',
                      'physical_propagation>=0.5.1']
)



import xml.etree.ElementTree
from .harmonic_series_representation import HarmonicSeries, HarmonicSpectrumNoiseSource


class InterpolatedHarmonicSpectrumNoiseSource:
    """
    We understand the data collected at Owens-Corning can construct a surface for the acoustic emissions on the surface
    of a sphere.  The Static_NoiseSource class represents a single engine power value for the definition of the source.
    This provides the user with the interface to determine the source description at specific engine power settings.

    20220330 - FSM - Added properties to average the reference conditions across the low and high power settings
    """

    def __init__(self, filename: str = None):
        """
        This constructs the source definition for the interpolated static source description of the acoustic emissions
        for SUAS.
        :param filename: str - the path to the file that contains the definition
        """

        self.definitions = list()
        self._low_power = None
        self._high_power = None
        self._max_power = None
        self._name = None
        self.desired_power_setting = None
        self.SOURCE_DESCRIPTION_ERROR = "The low and high power settings must be HarmonicNoiseSource"

        if filename is None:
            return

        self.filename = filename

        tree = xml.etree.ElementTree.ElementTree()
        tree.parse(filename)

        #   Get the parameters for the definition from the root's attributes

        root = tree.getroot()

        self._low_power = float(root.attrib['lo_definition'])
        self._high_power = float(root.attrib['hi_definition'])
        self._max_power = float(root.attrib['max_definition'])
        self._name = root.attrib['name']
        self.desired_power_setting = self._low_power

        for defs in root[0]:
            self.definitions.append(HarmonicSpectrumNoiseSource(defs))

    @property
    def source_name(self):
        return self._name

    @source_name.setter
    def source_name(self, value):
        self._name = value

    @property
    def low_power_setting(self):
        return self.low_power_description.aircraft_power_setting

    @property
    def high_power_setting(self):
        return self.high_power_description.aircraft_power_setting

    @property
    def maximum_power_setting(self):
        return self.maximum_power_description.aircraft_power_setting

    @property
    def low_power_engine_parameter(self):
        return self._low_power

    @low_power_engine_parameter.setter
    def low_power_engine_parameter(self, value):
        self._low_power = value

    @property
    def hi_power_engine_parameter(self):
        return self._high_power

    @hi_power_engine_parameter.setter
    def hi_power_engine_parameter(self, value):
        self._high_power = value

    @property
    def max_power_engine_parameter(self):
        return self._max_power

    @max_power_engine_parameter.setter
    def max_power_engine_parameter(self, value):
        self._max_power = value

    @property
    def low_power_description(self):
        tmp = None

        for definition in self.definitions:
            if definition.aircraft_power_setting == self._low_power:
                tmp = definition
                break

        if tmp is not None:
            return tmp
        else:
            raise ValueError("The low power setting conditions could not be found in the collection of definitions")

    @property
    def high_power_description(self):
        tmp = None

        for definition in self.definitions:
            if definition.aircraft_power_setting == self._high_power:
                tmp = definition
                break

        if tmp is not None:
            return tmp
        else:
            raise ValueError("The high power setting conditions could not be found in the collection of definitions")

    @property
    def maximum_power_description(self):
        tmp = None

        for definition in self.definitions:
            if definition.aircraft_power_setting == self._max_power:
                tmp = definition
                break

        if tmp is not None:
            return tmp
        else:
            raise ValueError("The maximum power setting conditions could not be found in the collection of definitions")

    @property
    def static_definitions(self):
        return self.definitions

    @static_definitions.setter
    def static_definitions(self, value):
        self.definitions = value

    @property
    def desired_engine_power(self):
        return self.desired_power_setting

    @property
    def ref_temperature(self):
        if not (isinstance(self.low_power_description, HarmonicSpectrumNoiseSource) and
                isinstance(self.high_power_description, HarmonicSpectrumNoiseSource)):
            return ValueError(self.SOURCE_DESCRIPTION_ERROR)
        return (self.low_power_description.ref_temperature + self.high_power_description.ref_temperature) / 2

    @property
    def ref_pressure(self):
        if not (isinstance(self.low_power_description, HarmonicSpectrumNoiseSource) and
                isinstance(self.high_power_description, HarmonicSpectrumNoiseSource)):
            return ValueError(self.SOURCE_DESCRIPTION_ERROR)
        return (self.low_power_description.ref_pressure + self.high_power_description.ref_pressure) / 2

    @property
    def ref_humidity(self):
        if not (isinstance(self.low_power_description, HarmonicSpectrumNoiseSource) and
                isinstance(self.high_power_description, HarmonicSpectrumNoiseSource)):
            return ValueError(self.SOURCE_DESCRIPTION_ERROR)
        return (self.low_power_description.ref_humidity + self.high_power_description.ref_humidity) / 2

    @property
    def ref_distance(self):
        if not (isinstance(self.low_power_description, HarmonicSpectrumNoiseSource) and
                isinstance(self.high_power_description, HarmonicSpectrumNoiseSource)):
            return ValueError(self.SOURCE_DESCRIPTION_ERROR)
        return (self.low_power_description.ref_measurement_distance +
                self.high_power_description.ref_measurement_distance) / 2

    @desired_engine_power.setter
    def desired_engine_power(self, value):
        import warnings

        if value < self.low_power_setting:
            self.desired_power_setting = self.low_power_setting

            warnings.warn("The desired engine power setting was below the minimum within the definition.  The value "
                          "was reset to the lowest value in the definition")
        elif value > self.maximum_power_setting:
            self.desired_power_setting = self.maximum_power_setting

            warnings.warn("The desired engine power setting was above the maximum within the definition.  The value "
                          "was reset to the maximum value in the definition")
        elif (value > self.high_power_setting) and (value < self.maximum_power_setting):
            self.desired_power_setting = self.maximum_power_setting

            warnings.warn("The value was above the interpolation range for this aircraft definition.  The source "
                          "description used will be the maximum definition.")
        else:
            self.desired_power_setting = value

    def predict(self, location):
        """
        This function calculates the sound pressure level at a specific location and returns the entire spectrum
        :param location: SphericalCoordinate
            it is either the specific location on the surface as defined in Cartesian or spherical coordinates or the
            azimuthal angle
        :return: double, array-like
            the sound pressure level @ all frequencies within the definition
        """
        import numpy as np

        if self.desired_power_setting > self.high_power_setting:
            return self.maximum_power_description.predict(location)
        else:
            #   Determine the weighting function or the two coefficients
            lo_definition = self.low_power_description
            hi_definition = self.high_power_description

            w_lo = 1 + (self.desired_power_setting - lo_definition.aircraft_power_setting) / \
                   (lo_definition.aircraft_power_setting - hi_definition.aircraft_power_setting)
            w_hi = 1 - w_lo

            #   Determine the locations for the determination of the harmonic coefficient matrix
            spl = np.zeros((len(self.definitions[0].series_coefficients), 1))

            for i in range(len(spl)):
                #   Build the matrix that we will use to determine the real values after creating the interpolated
                #   coefficients.
                b = HarmonicSeries.harmonic_matrix(
                    self.low_power_description.series_coefficients[i].order,
                    azimuthal=location.azimuthal,
                    polar=location.polar)

                #   Use the weights to determine the linear combination of the coefficients
                c_lo = self.low_power_description.series_coefficients[i]
                c_hi = self.high_power_description.series_coefficients[i]

                c = w_lo * c_lo.coefficients + w_hi * c_hi.coefficients

                spl[i] = (b.dot(c))[0].real

        return spl

    def save(self, filename: str):
        """
        This function creates the source file description using the information within the class.  It uses the functions
        from the static source to build the tree for the individual definitions.

        :param filename: string - the path to write the output
        """

        root = xml.etree.ElementTree.Element("interpolated_harmonic_series_definition")

        #   Build the interpolated source definition tree
        root.set("name", self.source_name)
        root.set("lo_definition", str(self.low_power_engine_parameter))
        root.set("hi_definition", str(self.hi_power_engine_parameter))
        if self._max_power is not None:
            root.set("max_definition", str(self._max_power))
        else:
            root.set("max_definition", str(self.hi_power_engine_parameter))

        definition_set = xml.etree.ElementTree.SubElement(root, "definitions")
        for definition in self.definitions:
            definition_set.append(self.definitions[i].m)

        #   Set the root of the tree and write the structure to the output file

        tree._setroot(interp_src)

        print(tree)
        tree.write(filename, encoding="UTF-8", xml_declaration=True)


import mne
import numpy as np
from pytimbre.waveform import Waveform
from PythonCoordinates.coordinates.coordinate_representations import CartesianCoordinate, SphericalCoordinate
from PythonCoordinates.measurables.physical_quantities import Length, Angle


class EEGEquivalentSource:
    """
    This class represents a conversion of the measured levels in the Electroencephalogram (EEG) files that are
    compatible with the MNE project. It will create a series of source levels at the scalp and fit this with a series
    of harmonic functions.
    """

    def __init__(self, filename: str):
        """
        This function loads the data from the file using the MNE module. A series of Waveform objects will be created
        to facilitate manipulation of the signals within PyTimbre.

        Parameters
        ----------

        :param filename: str
            The full path to the file that we are interested in loading.
        """

        self._eeg_data = None

        self._load_eeg_file(filename)

    def _load_eeg_file(self, filename: str):
        """
        This function will load the data from the binary file and create a series of Waveform objects that will be
        stored in the class.

        Parameters
        ----------

        :param filename: str
            The path to the file that we will load into the interface
        """

        #   Read the file
        self._eeg_data = mne.io.read_raw_edf(filename, preload=True, verbose=0)

        #   Set the reference of the EEG to the mastoids and apply a window to the data
        self._eeg_data.set_eeg_reference(['M1', 'M2'])
        self._eeg_data.filter(1., 40., fir_design='firwin')

    @staticmethod
    def raw_cartesian_location(electrode_index: int = 0, system_name: str = "biosemi64"):
        """
        Using the index of the electrode, query the system and determine the location of the electrode and return
        the location in the CartesianCoordinate object.

        Parameters
        ----------
        :param electrode_index: int
            The index within the system of the cap that we want to query
        :param system_name: str
            The name of the system that exists within the definition of the systems within MNE.
        :return: CartesianCoordinate
            The location from the montage within the MNE module
        """

        #   Get the montage based on the system name - this is an internal representation of the data
        montage = mne.channels.make_standard_montage(system_name)
        location = montage.dig[electrode_index]['r']

        #   Create the location information for the electrode
        return CartesianCoordinate(Length(location[0]),
                                   Length(location[1]),
                                   Length(location[2]))

    @staticmethod
    def electrode_name(electrode_index: int = 0, system_name: str = "biosemi64"):
        """
        This will query the montage within the system and provide the name of the electrode that is queried.

        Parameters
        ----------
         :param electrode_index: int
            The index within the system of the cap that we want to query
        :param system_name: str
            The name of the system that exists within the definition of the systems within MNE.
        :return: str
            The name of the electrode with the selected index
        """

        return mne.channels.make_standard_montage(system_name).ch_names[electrode_index]

    @property
    def sample_rate(self):
        return self._eeg_data.info["sfreq"]

    @property
    def time_data(self, channel_index: int = 0):
        return self._eeg_data._data[channel_index, :]

    @property
    def duration(self):
        return self._eeg_data.tmax

    def get_waveform(self, electrode_index: int = 0, t0: float = 10, t1: int = None) -> Waveform:
        """
        This will extract the information from the electrode within the file that was loaded and return the temporal
        data as a waveform.

        Parameters
        ----------
        :param electrode_index:
            The index within the data for the channel that we want to extract - see raw_cartesian_location and
            electrode_name
        :param t0:
            The start time of the waveform to obtain from the file
        :param t1:
            The end time of the waveform
        :return:
            A Waveform object that we can manipulate with PyTimbre
        """

        #   Get the samples that we need to extract based on the time window
        s0 = int(np.floor(t0 * self.sample_rate))

        #  Set the end sample
        if t1 is not None:
            s1 = int(np.floor(t1 * self.sample_rate))
        else:
            s1 = self._eeg_data.last_samp

        samples = self._eeg_data._data[electrode_index, s0:s1]

        return Waveform(samples, self.sample_rate, 0)










